#include "SiriSSL.h"

#define EVP_PKEY_assign_rsa     EVP_PKEY_assign_RSA
#define EVP_PKEY_assign_dh      EVP_PKEY_assign_DH
#define EVP_PKEY_assign_dsa     EVP_PKEY_assign_DSA

MODULE = SiriSSL::EVP_PKEY      PACKAGE = SiriSSL::EVP_PKEY   PREFIX = EVP_PKEY


int
EVP_PKEY_assign_rsa(pkey, rsa)
        EVP_PKEY *      pkey
        RSA *           rsa

int
EVP_PKEY_assign_dsa(pkey, dsa)
        EVP_PKEY *      pkey
        DSA *           dsa

int
EVP_PKEY_assign_dh(pkey, dh)
        EVP_PKEY *      pkey
        DH *            dh

int
EVP_PKEY_bits(pkey)
        EVP_PKEY *pkey

int        
EVP_PKEY_size(pkey)
        EVP_PKEY *pkey

SV *
EVP_PKEY_encrypt(pkey, plaintext)
                EVP_PKEY *      pkey
                SV *            plaintext
        PREINIT:
                SV *    ciphertext;
                int     p_len;
                char *  p_cp;
        CODE:
                if(SvPOK(plaintext))
                        p_cp = SvPV(plaintext, p_len);
                else
                        p_cp = SvPV_force(sv_2mortal(newSVsv(plaintext)),
                                          p_len);

                SV_ALLOC(ciphertext, p_len);
                EVP_PKEY_encrypt(SvPVX(ciphertext), p_cp, p_len, pkey);

                RETVAL = ciphertext;
        OUTPUT:
                RETVAL

        
SV *
EVP_PKEY_decrypt(pkey, ciphertext)
                EVP_PKEY *      pkey
                SV *            ciphertext
        PREINIT:
                SV *    plaintext;
                int     c_len;
                char *  c_cp;
        CODE:
                if(SvPOK(ciphertext))
                        c_cp = SvPV(ciphertext, c_len);
                else
                        c_cp = SvPV_force(sv_2mortal(newSVsv(ciphertext)),
                                          c_len);

                SV_ALLOC(plaintext, c_len);
                EVP_PKEY_decrypt(SvPVX(plaintext), c_cp, c_len, pkey);

                RETVAL = plaintext;
        OUTPUT:
                RETVAL

void
EVP_PKEY_DESTROY(...)
        PREINIT:
                EVP_PKEY *pkey;
        CODE:
                CLASS_IN(ST(0), pkey, EVP_PKEY *, SiriSSL::EVP_PKEY);
                printd("destroying EVP_PKEY object (0x%x): %d references\n",
                       (unsigned long)pkey, pkey->references);
                EVP_PKEY_free(pkey);


