#include "SiriSSL.h"
#include <pem.h>

#define PEM_set_passwd_callback         GLOBAL_set_perl_passwd_callback

#define XPUSH_PEM_DECODE(class, pem_surfix, sv) \
        do { \
                void *object; \
                BIO *bio; \
                I32 gimme = GIMME_V; \
                \
                if(gimme == G_VOID) \
                        break; \
                \
                bio = BIO_new(BIO_s_mem()); \
                BIO_write(bio, SvPVX(sv), SvCUR(sv)); \
                object = PEM_read_bio_ ## pem_surfix (bio, NULL, \
                                                      PEM_PASSWD_CB); \
                if(object) \
                        XPUSH_OBJ(object, class); \
                else \
                        goto end; \
                \
                if(gimme != G_ARRAY) \
                        goto end; \
                \
                while(object = PEM_read_bio_ ## pem_surfix (bio, NULL, \
                                                            PEM_PASSWD_CB)) \
                        XPUSH_OBJ(object, class); \
                \
                end: \
                BIO_free(bio); \
        } while(0)

#define XPUSH_PEM_ENCODE(pem_surfix, object) \
        do { \
                BIO *bio; \
                BUF_MEM *bm; \
                \
                bio = BIO_new(BIO_s_mem()); \
                if(PEM_write_bio_ ## pem_surfix (bio, object)) { \
                        bm = (BUF_MEM *)bio->ptr; \
                        XPUSHs(sv_2mortal(newSVpv(bm->data, bm->length))); \
                } else { \
                        XPUSHs(&sv_undef); \
                } \
                \
                BIO_free(bio); \
        } while(0)


#define XPUSH_PEM_ENCODE_SECRET(pem_surfix, object) \
        do { \
                BIO *           bio; \
                BUF_MEM *       bm; \
                EVP_CIPHER *    cipher = NULL; \
                \
                if(items > 1) \
                        cipher = __get_cipher(ST(1)); \
                \
                bio = BIO_new(BIO_s_mem()); \
                if(PEM_write_bio_ ## pem_surfix (bio, object, cipher, \
                                                 NULL, 0, PEM_PASSWD_CB)) { \
                        bm = (BUF_MEM *)bio->ptr; \
                        XPUSHs(sv_2mortal(newSVpv(bm->data, bm->length))); \
                } else { \
                        XPUSHs(&sv_undef); \
                } \
                \
                BIO_free(bio); \
        } while(0)

#define PEM_PASSWD_CB    \
        (GLOBAL_get_perl_passwd_callback(0) ? \
         perl_passwd_callback : NULL)

EVP_CIPHER *__get_cipher(SV *cipher)
{
        char *cp;

        if(SvIOK(cipher)) {
                if(SvIV(cipher))
                        return PEM_DEFAULT_CIPHER;
                else
                        return NULL;
        } else if(SvPOK(cipher)) {
                cp = SvPVX(cipher);
                return EVP_get_cipherbyname(cp);
        }
        return NULL;
}

MODULE = SiriSSL::PEM   PACKAGE = SiriSSL::PEM  PREFIX = PEM_

void
PEM_set_passwd_callback(callback)
                SV *callback

void
PEM__decode_cert(sv_buf)
                SV *sv_buf                
        PPCODE:
                XPUSH_PEM_DECODE(SiriSSL::X509, X509, sv_buf);

void
PEM__decode_crl(sv_buf)
                SV *sv_buf
        PPCODE:
                XPUSH_PEM_DECODE(SiriSSL::X509_CRL, X509_CRL, sv_buf);

void
PEM__decode_req(sv_buf)
                SV *sv_buf
        PPCODE:
                XPUSH_PEM_DECODE(SiriSSL::X509_REQ, X509_REQ, sv_buf);

void
PEM_decode_secret_key(sv_buf)
                SV *sv_buf
        PPCODE:
                XPUSH_PEM_DECODE(SiriSSL::EVP_PKEY, PrivateKey, sv_buf);

void
PEM__decode_secret_rsa_key(sv_buf)
                SV *sv_buf
        PPCODE:
                XPUSH_PEM_DECODE(SiriSSL::RSA, RSAPrivateKey, sv_buf);

void
PEM__decode_secret_dsa_key(sv_buf)
                SV *sv_buf
        PPCODE:
                XPUSH_PEM_DECODE(SiriSSL::DSA, DSAPrivateKey, sv_buf);


void
PEM__encode_cert(x509)
                X509 *x509
        PPCODE:
                XPUSH_PEM_ENCODE(X509, x509);                

void
PEM__encode_crl(crl)
                X509_CRL *crl
        PPCODE:
                XPUSH_PEM_ENCODE(X509_CRL, crl);

void
PEM__encode_req(req)
                X509_REQ *req
        PPCODE:
                XPUSH_PEM_ENCODE(X509_REQ, req);

void
PEM_encode_secret_key(key, ...)
                EVP_PKEY *key
        PPCODE:
                XPUSH_PEM_ENCODE_SECRET(PrivateKey, key);
                
void
PEM_encode_secret_rsa_key(key, ...)
                RSA *key
        PPCODE:
                XPUSH_PEM_ENCODE_SECRET(RSAPrivateKey, key);

void
PEM_encode_secret_dsa_key(key, ...)
                DSA *key
        PPCODE:
                XPUSH_PEM_ENCODE_SECRET(DSAPrivateKey, key);
