#include "SiriSSL.h"

SSL_METHOD *
__get_ssl_method(method)
        SV *method;
{
        switch(SvIV(method)) {

        case SSL_METHOD_SSLv23:
                return SSLv23_method();

        case SSL_METHOD_SSLv3:
                return SSLv3_method();

        case SSL_METHOD_SSLv2:
                return SSLv2_method();
        default:
                croak("Unknown SSL_METHOD type");
        }

        return NULL;
}

#define X509_STORE__del_cert         X509_STORE_del_cert

static int
X509_STORE__del_cert(X509_STORE *store, X509 *cert)
{
        X509_OBJECT obj, *tmp;

        obj.type = X509_LU_X509;
        obj.data.x509 = cert;

        tmp = (X509_OBJECT *)lh_delete(store->certs, (char *)&obj);
        if(!tmp)
                return 0;

        X509_OBJECT_free_contents(tmp);
        return 1;
}

MODULE = SiriSSL::SSL_CTX  PACKAGE = SiriSSL::SSL_CTX     PREFIX = SSL_CTX_

SSL_CTX *
SSL_CTX_new(class, ...)
        SV *class

        PREINIT:
        
        SSL_METHOD *ssl_method;

        CODE:
                if(items > 1) ssl_method = __get_ssl_method(ST(1));
                else ssl_method = SSL_METHOD_DEFAULT;

                RETVAL = SSL_CTX_new(ssl_method);
        OUTPUT:
                RETVAL

void
SSL_CTX_DESTROY(...)
        PREINIT:
                SSL_CTX *ctx;
        CODE:
                CLASS_IN(ST(0), ctx, SSL_CTX *, SiriSSL::SSL_CTX);
                printd("Destroying ssl_ctx perl object (0x%x): %d references\n",
                       (unsigned long)ctx, ctx->references);
                SSL_CTX_free(ctx);


int
SSL_CTX_use_RSAPrivateKey(ctx, rsa_key)
        SSL_CTX *ctx
        RSA *rsa_key

int
SSL_CTX_use_PrivateKey(ctx, pkey)
        SSL_CTX *ctx
        EVP_PKEY *pkey

int
SSL_CTX_use_certificate(ctx, x509)
        SSL_CTX *ctx
        X509 *x509

int
SSL_CTX_use_RSAPrivateKey_file(ctx, file, type = X509_FILETYPE_PEM)
        SSL_CTX *ctx
        char *file
        int type

int
SSL_CTX_use_PrivateKey_file(ctx, file, type = X509_FILETYPE_PEM)
        SSL_CTX *ctx
        char *file
        int type

int
SSL_CTX_use_certificate_file(ctx, file, type = X509_FILETYPE_PEM)
        SSL_CTX *ctx
        char *file
        int type

int
SSL_CTX_check_private_key(ctx)
        SSL_CTX *ctx

long
SSL_CTX_set_options(ctx, options)
        SSL_CTX *ctx
        int options


long
SSL_CTX_get_options(ctx)
        SSL_CTX *ctx

int
SSL_CTX_set_cipher_list(ctx, cipher_list)
        SSL_CTX *ctx
        char *cipher_list

int
SSL_CTX_load_verify_locations(ctx, CAfile, CApath)
        SSL_CTX *ctx
        char *CAfile
        char *CApath


int
SSL_CTX_set_default_verify_paths(ctx)
        SSL_CTX *ctx


void
SSL_CTX_set_verify(ctx, flags, sv_callback)
        SSL_CTX *       ctx
        int             flags
        SV *            sv_callback

        PREINIT:
                X509_STORE *cert_store;
        CODE:
                cert_store = SSL_CTX_get_cert_store(ctx);
                X509_STORE_set_perl_verify_callback(cert_store, sv_callback);
                
                SSL_CTX_set_verify(ctx, flags, perl_verify_callback);

void
SSL_CTX_set_tmp_rsa_callback(ctx, sv_callback)
        SSL_CTX *       ctx
        SV *            sv_callback

        CODE:
                SSL_CTX_set_perl_tmp_rsa_callback(ctx, sv_callback);
                SSL_CTX_set_tmp_rsa_callback(ctx, perl_tmp_rsa_callback);

void
SSL_CTX_set_tmp_dh_callback(ctx, sv_callback)
        SSL_CTX *       ctx
        SV *            sv_callback

        CODE:
                SSL_CTX_set_perl_tmp_dh_callback(ctx, sv_callback);
                SSL_CTX_set_tmp_dh_callback(ctx, perl_tmp_dh_callback);


void
SSL_CTX_set_info_callback(ctx, sv_callback)
        SSL_CTX *       ctx
        SV *            sv_callback

        CODE:
                SSL_CTX_set_perl_info_callback(ctx, sv_callback);
                SSL_CTX_set_info_callback(ctx, perl_info_callback);


void
SSL_CTX_set_passwd_callback(ctx, sv_callback)
        SSL_CTX *       ctx
        SV *            sv_callback

        CODE:
                GLOBAL_set_perl_passwd_callback(sv_callback);
                SSL_CTX_set_default_passwd_cb(ctx, perl_passwd_callback);


SV *
SSL_CTX__trusted_certs(ctx, ...)
                SSL_CTX *ctx
        PREINIT:
                SV *cert_store;
        CODE:
                cert_store = SSL_CTX_get_perl_cert_store(ctx);
                if(items > 1) {
                        cert_store = ST(1);
                        CertStore_parent(cert_store, ctx);
                        SSL_CTX_set_perl_cert_store(ctx, cert_store);
                } else if(!cert_store) {
                        cert_store = CertStore_new(ctx);
                        SSL_CTX_set_perl_cert_store(ctx, cert_store);
                }
                RETVAL = newSVsv(cert_store);
        OUTPUT:
                RETVAL


int
SSL_CTX__add_trusted_certs(ctx, ...)
                SSL_CTX *       ctx
        PREINIT:
                int i, count = 0;

                X509_STORE *    store;
                X509 *          cert;
        CODE:
                store = SSL_CTX_get_cert_store(ctx);

                for(i = 1; i < items; i++) {
                        CLASS_IN(ST(i), cert, X509 *, SiriSSL::X509);
                        count += X509_STORE_add_cert(store, cert);
                }
                RETVAL = count;                
        OUTPUT:
                RETVAL


int
SSL_CTX__del_trusted_certs(ctx, ...)
                SSL_CTX *       ctx
        PREINIT:
                int i, count = 0;

                X509_STORE *    store;
                X509 *          cert;
        CODE:
                store = SSL_CTX_get_cert_store(ctx);

                for(i = 1; i < items; i++) {
                        CLASS_IN(ST(i), cert, X509 *, SiriSSL::X509);
                        count += X509_STORE_del_cert(store, cert);
                }
                RETVAL = count;
        OUTPUT:
                RETVAL
