#ifdef __cplusplus
#extern "C" {
#endif

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#ifdef __cplusplus
}
#endif

#include <ssl.h>

#define SSL_DEBUG
#ifdef SSL_DEBUG
#define printd(args...)         printf(##args)
#else
#define printd(args...)
#endif

#define MATCH_CONSTANT(name, prefix) \
        if(strnEQ(name, #prefix, strlen(#prefix))) { \
                double value = prefix ## CONSTANT(name + strlen(#prefix)); \
                if(!errno) \
                        return value; \
        }

#define InOutStream     PerlIO *

#define CLASS_IN(arg, var, type, class) \
        if(sv_isobject(arg) && sv_derived_from(arg, #class)) \
                var = (type) SvIV((SV *)SvRV(arg)); \
        else \
                croak(#var " is not " #class);

#define CLASS_OUT(arg, var, type, class) \
                sv_setref_pv(arg, #class, (PV *)var);

#define CLASS_OUT_GC(arg, var, type, class) \
        do { \
                REF_INC(var, type); \
                sv_setref_pv(arg, #class, (PV *)var); \
        } \
        while(0);

#define XPUSH_OBJ(ptr, class) \
                XPUSHs(sv_setref_pv(sv_newmortal(), #class, (PV *)ptr))

#define XPUSH_NV(nv) \
                XPUSHs(sv_2mortal(newSVnv(nv)))

#define XPUSH_IV(iv) \
                XPUSHs(sv_2mortal(newSViv(iv)))

#define XPUSH_PV(cp, len) \
                XPUSHs(sv_2mortal(newSVpv(cp, len)))

#define POP_OBJ(class) \
        sv_unref_ptr(POPs, #class)
        
#define BOOT_DEF(OBJ)   extern void boot_SiriSSL__ ## OBJ(CV *cv)
#define BOOT_CALL(OBJ)  callXS(boot_SiriSSL__ ## OBJ , cv, mark)

/* useful perl SV macros */
#define SV_ALLOC(sv, len) \
        do { \
                sv = newSVpv("", 0); \
                SvGROW(sv, len); \
        } while(0)

#define SV_SETSV(a, b) \
        do { \
                if(a == (SV *)NULL) \
                        a = newSVsv(b); \
                else \
                        SvSetSV(a, b); \
        } while(0)

#define SV_SETIV(sv, iv) \
        do { \
                if(sv == (SV *)NULL) \
                        sv = newSViv(iv); \
                else \
                        sv_setiv(sv, iv); \
        } while(0)

#define EXTENSION_STORE(obj, idx, arg) \
        ex_store(&obj->ex_data, #idx, idx, (void *)arg)

#define EXTENSION_DEMAND(obj, idx) \
        ex_fetch(&obj->ex_data, #idx, idx, 1)

#define EXTENSION_FETCH(obj, idx) \
        ex_fetch(&obj->ex_data, #idx, idx, 0)

#define DIGEST_MAX_LEN          64
#define ERR_MAX_LEN             512
#define CIPHERS_MAX_LEN         1024

/**************************
 *** SSL API extensions ***
 **************************/

#define SSL_METHOD_SSLv2        2
#define SSL_METHOD_SSLv23       23
#define SSL_METHOD_SSLv3        3

#define SSL_ERROR_CLOSED        SSL_ERROR_ZERO_RETURN

#define SSL_METHOD_DEFAULT      SSLv23_method()   /* More flexible */

#define OR(a, b)                ((a) ? (a) : (b))
#define EVP(cipher)             EVP_get_cipherbyname( #cipher )
#define MD(digest)              EVP_get_digestbyname( #digest )

#define PEM_DEFAULT_CIPHER      OR(EVP(des3), \
                                OR(EVP(idea), \
                                OR(EVP(blowfish), \
                                OR(EVP(desx), \
                                OR(EVP(cast), \
                                OR(EVP(des), EVP(rc5)))))))

#define X509_DEFAULT_DIGEST     OR(MD(sha), \
                                OR(MD(md5), \
                                OR(MD(sha1), \
                                OR(MD(ripemd160), \
                                OR(MD(mdc2), MD(md2))))))

/*#define SSL_METHOD_DEFAULT       SSLv3_method()  /* More secure */

#define REF_ADD(obj, n, type) \
        CRYPTO_add(&obj->references, n, CRYPTO_LOCK_ ## type )

#define REF_INC(obj, type)      REF_ADD(obj, 1, type)
#define REF_DEC(obj, type)      REF_ADD(obj, -1, type)

#define REF_CNT(obj)            (obj->references)
#define REF_ISLAST(obj)         (obj->references == 1)

SV *    CertStore_new(SSL_CTX *parent);
void    CertStore_parent(SV *cert_store, SSL_CTX *parent);

/*
        Object types to watch for reference increments:
        X509 X509_INFO X509_PKEY X509_CRL X509_REQ
        X509_STORE

        DSA RSA EVP_PKEY

        SSL_CTX SSL_CERT SSL_SESSION
        SSL BIO
*/

unsigned long   SSL_errno(void);
char *          SSL_strerror(unsigned long SSL_errno);

void    X509_STORE_set_perl_verify_callback(X509_STORE *cert_store, SV *cb);
SV *    X509_STORE_get_perl_verify_callback(X509_STORE *cert_store);

void    SSL_set_io_retval(SSL *ssl, int io_retval);
int     SSL_get_io_retval(SSL *ssl);

/* TODO: add extension in a civil manner using ex_new_index, registering
         a proper destructor which decrements our object and destroys it
         when the rest of SSL_CTX is finally destroyed. */
void    SSL_CTX_set_perl_cert_store(SSL_CTX *ssl_ctx, SV *cert_store);
SV *    SSL_CTX_get_perl_cert_store(SSL_CTX *ssl_ctx);

void    SSL_CTX_set_perl_tmp_rsa_callback(SSL_CTX *ssl_ctx, SV *cb);
SV *    SSL_CTX_get_perl_tmp_rsa_callback(SSL_CTX *ssl_ctx);

void    SSL_CTX_set_perl_tmp_dh_callback(SSL_CTX *ssl_ctx, SV *cb);
SV *    SSL_CTX_get_perl_tmp_dh_callback(SSL_CTX *ssl_ctx);

void    SSL_CTX_set_perl_info_callback(SSL_CTX *ssl_ctx, SV *cb);
SV *    SSL_CTX_get_perl_info_callback(SSL_CTX *ssl_ctx);

void    GLOBAL_set_perl_passwd_callback(SV *cb);
SV *    GLOBAL_get_perl_passwd_callback();

/*****************************************
 *** Callback C => Perl glue functions ***
 *****************************************/

int     perl_verify_callback(int ok, X509_STORE_CTX *store_ctx);
void    perl_info_callback(SSL *ssl, int where, int ret);
RSA *   perl_tmp_rsa_callback(SSL *ssl, int export, int keylen);
DH *    perl_tmp_dh_callback(SSL *ssl, int export, int keylen);
int     perl_passwd_callback(char *buf, int buf_len, int verify);
