# SiriSSL.pm
#
# Author: Liraz Siri <liraz@hushmail.com>
#
# Copyright (c) 1999 Liraz Siri <liraz@hushmail.com>, Ariel, Israel
#               All rights reserved.
#
# Created: Sun Jun 20 05:22:15 IDT 1999

package SiriSSL;
use DynaLoader;
use Exporter;

use strict;
use Carp;
use vars qw(@ISA $VERSION @EXPORT_OK $AUTOLOAD %SYMBOLS %EXPORT_TAGS);

$VERSION = '0.01';
@ISA = qw(DynaLoader Exporter);

# soon will be obsolete.
%SYMBOLS = (CB     => [qw(SSL_CB_ALERT
                          SSL_CB_EXIT
                          SSL_CB_LOOP
                          SSL_CB_READ
                          SSL_CB_WRITE)],

            ERROR  => [qw(SSL_ERROR_NONE
                          SSL_ERROR_SSL
                          SSL_ERROR_SYSCALL
                          SSL_ERROR_WANT_CONNECT
                          SSL_ERROR_WANT_READ
                          SSL_ERROR_WANT_WRITE
                          SSL_ERROR_WANT_X509_LOOKUP
                          SSL_ERROR_ZERO_RETURN)],

            OP     => [qw(SSL_OP_ALL
                          SSL_OP_NETSCAPE_DEMO_CIPHER_CHANGE_BUG
                          SSL_OP_NON_EXPORT_FIRST
                          SSL_OP_NO_SSLv2
                          SSL_OP_NO_SSLv3
                          SSL_OP_NO_TLSv1)],

            SESS   => [qw(SSL_SESS_CACHE_BOTH
                          SSL_SESS_CACHE_CLIENT
                          SSL_SESS_CACHE_NO_AUTO_CLEAR
                          SSL_SESS_CACHE_NO_INTERNAL_LOOKUP
                          SSL_SESS_CACHE_OFF
                          SSL_SESS_CACHE_SERVER)],

            ST     => [qw(SSL_ST_ACCEPT
                          SSL_ST_CONNECT
                          SSL_ST_MASK)],

            METHOD => [qw(SSL_METHOD_SSLv2
                          SSL_METHOD_SSLv23
                          SSL_METHOD_SSLv3)],
            
            VERIFY => [qw(SSL_VERIFY_CLIENT_ONCE
                          SSL_VERIFY_FAIL_IF_NO_PEER_CERT
                          SSL_VERIFY_NONE
                          SSL_VERIFY_PEER)],

            X509 => [qw(X509_V_OK
                        X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT
                        X509_V_ERR_UNABLE_TO_GET_CRL
                        X509_V_ERR_UNABLE_TO_DECRYPT_CERT_SIGNATURE
                        X509_V_ERR_UNABLE_TO_DECRYPT_CRL_SIGNATURE
                        X509_V_ERR_UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY
                        X509_V_ERR_CERT_SIGNATURE_FAILURE
                        X509_V_ERR_CRL_SIGNATURE_FAILURE
                        X509_V_ERR_CERT_NOT_YET_VALID
                        X509_V_ERR_CERT_HAS_EXPIRED
                        X509_V_ERR_CRL_NOT_YET_VALID
                        X509_V_ERR_CRL_HAS_EXPIRED
                        X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD
                        X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD
                        X509_V_ERR_ERROR_IN_CRL_LAST_UPDATE_FIELD
                        X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD
                        X509_V_ERR_OUT_OF_MEM
                        X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
                        X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN
                        X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY
                        X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE
                        X509_V_ERR_CERT_CHAIN_TOO_LONG
                        X509_V_ERR_CERT_REVOKED
                        X509_V_ERR_APPLICATION_VERIFICATION)]);

@EXPORT_OK = (map(@$_, values %SYMBOLS));
%EXPORT_TAGS = (error => $SYMBOLS{ERROR},
                verify => $SYMBOLS{VERIFY},
                state => $SYMBOLS{ST},
                sess => $SYMBOLS{SESS},
                options => $SYMBOLS{OP},
                callback => $SYMBOLS{CB},
                method => $SYMBOLS{METHOD},
                x509 => $SYMBOLS{X509},
                all => \@EXPORT_OK);

sub AUTOLOAD {
    my $constname;
    ($constname = $AUTOLOAD) =~ s/.*:://;

    croak "Undefined subroutine &$AUTOLOAD called"
        if($constname eq 'constant' || @_);

    my $val = constant($constname);

    croak "Undefined subroutine &$AUTOLOAD called" if ($! != 0);

    eval "sub $AUTOLOAD { $val };";
    goto &$AUTOLOAD;
}

bootstrap SiriSSL $VERSION;

package SiriSSL::SSL;

use UNIVERSAL qw(isa);

sub peer_certified_by {
    my $self = shift; my $arg = shift;

    if(isa($arg, 'SiriSSL::X509') || ref($arg) eq 'ARRAY') {
        return $self->peer_certified_by_certs($arg, @_);
    } else {
        return $self->peer_certified_by_name($arg, @_);
    }
    
    return 0;
}

sub peer_certified_by_certs {
    my ($self, $arg, $depth_limit) = @_;
    my @chain = $self->peer_cert_chain;

    my @trusted_certs = ref($arg) eq 'ARRAY' ? @$arg : ($arg);

    my $i = 0;
    while(($i < @chain) && ($i <= $depth_limit)) {
        return 1 if(grep($chain[$i] eq $_, @trusted_certs));
        $i++;
    }
    return 0;
}

sub peer_certified_by_name {
    my ($self, $arg, $depth_limit) = @_;
    my @chain = $self->peer_cert_chain;

    my $name = (isa($arg, 'SiriSSL::X509_NAME') ?
                $arg : SiriSSL::X509_NAME->new($arg));
    my $i = 0;
    while(($i < @chain) && ($i <= $depth_limit)) {
        return 1 if($chain[$i]->name->match($name) &&
                    $self->ctx->trusted_certs->fetch($chain[$i]));
        $i++;
    }
    return 0;
}

sub print {
    my $self = shift;
    my $msg = @_ ? join('', @_) : $_;

    return defined $self->write($msg) ? 1 : undef;
}

sub printf {
    my $self = shift;

    my $msg = sprintf(shift @_, @_);
    $self->print($msg);
}

sub write {
    my $self = shift;
    my $msg = shift;
    my $length = @_ ? shift : length($msg);
    my $offset = @_ ? shift : 0;

    my $count = 0;
    while($length - $count > 0) {
        my $n = $self->syswrite($msg, $length - $count, $offset + $count);
        return ($count ? $count : undef) unless $n;

        $count += $n;
    }

    return $count;
}

sub read {
    my $self = shift;
    my @args = @_[1 .. $#_];

    return $self->sysread($_[0], 1024) unless(@args);

    my ($length, $offset) = @args;
    $offset ||= 0;
    
    my $count = 0;
    while($length - $count> 0) {
        my $n = $self->sysread($_[0], $length - $count, $offset + $count);
        return ($count ? $count : undef) unless $n;

        $count += $n;
    }
    return $count;
}

sub error_str {
    my $self = shift;
    my $e = $self->error;

    no strict 'refs';

    my ($NONE, $SSL, $SYSCALL, $WANT_READ, $WANT_WRITE, $WANT_X509_LOOKUP,
        $ZERO_RETURN, $WANT_CONNECT) =
            map(&{"SiriSSL::SSL_ERROR_$_"},
                qw[NONE SSL SYSCALL WANT_READ WANT_WRITE WANT_X509_LOOKUP
                   ZERO_RETURN WANT_CONNECT]);

    return ($e eq $NONE ? '' :
            $e eq $ZERO_RETURN ? 'connection closed' :
            
            $e eq $SYSCALL ? "syscall: $!" :
            $e eq $SSL ? sprintf('ssl: %s', SiriSSL::ERR::reason()) :

            $e eq $WANT_READ ? 'want read' :
            $e eq $WANT_WRITE ? 'want write' :
            $e eq $WANT_CONNECT ? 'want connect' :

            $e eq $WANT_X509_LOOKUP ? 'want x509 lookup' :

            undef);
}

package SiriSSL::SSL_CTX;

sub trusted_certs {
    my $self = shift;
    my $cert_store = $self->_trusted_certs;
    
    if(@_) {
        my $arg = shift;

        if(UNIVERSAL::isa($arg, ref $cert_store)) {
            $cert_store->parent(undef);
            $cert_store = $arg;
            $self->_trusted_certs($cert_store);
        } else {
            $cert_store->store($arg, @_);
        }
    }
    $cert_store;
}

package SiriSSL::PEM;

use IO::File;
use vars qw($AUTOLOAD);

sub AUTOLOAD {
    my ($class, $method) = ($AUTOLOAD =~ /^(.*)::([^:]*)$/);

    my $alias = ($method =~ s/s$//);
    my $internal_method = "_$method";  # decode_certs() == decode_cert()

    die "Undefined subroutine &$AUTOLOAD called"
        unless $class->can($internal_method);

    if($alias) {
        no strict 'refs';
        *{$AUTOLOAD} = \&{$class . '::' . $method};
        goto &$AUTOLOAD if($class->can($method));
    }
    
    ($method =~ /^decode_/) && eval<<EOF;
package $class;
    
sub $method {    
    my \@objs;
    for my \$arg (\@_) {
        push(\@objs, \$arg), next if(ref(\$arg) =~ /^SiriSSL::/);
        if(length(\$arg) < 128 && stat(\$arg)) {
            my \$fh = IO::File->new("< \$arg")
                or next;
            push \@objs, $internal_method(join('', <\$fh>));
            \$fh->close;
        } else {
            push \@objs, $internal_method(\$arg);
        }
    }
    return wantarray ? \@objs : \$objs[0];
}
EOF

        ($method =~ /^encode_/) && eval<<EOF;
package $class;

sub $method {
    if(ref(\$_[0]) eq 'ARRAY') {
        my \@lol;
        foreach my \$arg (\@_) {
            push \@lol, $method(ref(\$arg) eq 'ARRAY' ? \@\$arg : \$arg);
        }
        return wantarray ? \@lol : \$lol[0];
    } else {
        my \@encoded;
        foreach my \$obj (\@_) {
            my \$encoded = \$obj->can('encode') ?
                \$obj->encode : $internal_method(\$obj);
        
            push \@encoded, \$encoded;
        }
        return join("\n", \@encoded);
    }
}
EOF
    goto &$AUTOLOAD;
}

package SiriSSL::X509;
use Carp;
use overload
    '""' => sub { SiriSSL::PEM::encode_cert(shift @_) },
    'eq' => sub { !$_[0]->cmp($_[1]) },
    'cmp' => sub { $_[0]->cmp($_[1]) },
    fallback => 1;

use UNIVERSAL qw(isa);
use vars qw(%ENCODE_CLEARTEXT_DEFAULTS
            @ENCODE_CLEARTEXT_TOKENS
            %ENCODE_CLEARTEXT_TOKEN_ALIASES);

@ENCODE_CLEARTEXT_TOKENS = qw(name issuer_name validity fingerprint bits);

%ENCODE_CLEARTEXT_DEFAULTS = (
    tokens => [ @ENCODE_CLEARTEXT_TOKENS ],
    formats => { names => 'shortline',
                 fingerprint => 'md5' },
    );

%ENCODE_CLEARTEXT_TOKEN_ALIASES = (
    names => ['name', 'issuer_name'],
    all => \@ENCODE_CLEARTEXT_TOKENS,
    default => $ENCODE_CLEARTEXT_DEFAULTS{tokens},
    expires => ['validity'],
    issuer => ['issuer_name'],
    none => []
    );

sub encode_defaults {
    my $self = shift;

    if(ref($_[0]) eq 'HASH') {
        my $formats = shift;
        foreach my $key (keys %$formats) {
            my $value = $formats->{$key};
            die "Unknown '$key' formatter requested: $value"
                if($key eq 'names' && !'SiriSSL::X509_NAME'->can($value));
            $ENCODE_CLEARTEXT_DEFAULTS{'formats'}{$key} = $value;
        }
    }
    
    if(@_) {
        my $aliases = \%ENCODE_CLEARTEXT_TOKEN_ALIASES;
        my @tokens = map(exists $aliases->{$_} ? @{$aliases->{$_}} : $_, @_);

        @{$ENCODE_CLEARTEXT_DEFAULTS{tokens}} = @tokens;
    }
    1;
}

sub encode {
    my $self = shift;

    my $splice_dups = sub {
        my $i = 0;
        while($i < @_) {
            splice(@_, $i, 1), next if($_[$i] eq $_[$i+1]);
            $i++;
        }
        @_;
    };
    
    my $i = 0;
    my %order = map {$_ => $i++} @ENCODE_CLEARTEXT_TOKENS;

    my %formats = %{$ENCODE_CLEARTEXT_DEFAULTS{'formats'}};
    %formats = (%formats, %{shift @_}) if(ref($_[0]) eq 'HASH');

    my @args = @_ ? @_ : ('default');

    my $aliases = \%ENCODE_CLEARTEXT_TOKEN_ALIASES;
    @args = map(exists $aliases->{$_} ? @{$aliases->{$_}} : $_, @args);
    @args = sort {$order{$a} <=> $order{$b}} grep(exists $order{$_}, @args);
    @args = $splice_dups->(@args);

    my $encoded;
    my $nameline = $formats{'names'} || 'shortline';

    foreach my $arg (@args) {
        if($arg eq 'name') {
            $encoded .= sprintf("Name: %s\n", $self->name->$nameline());
        } elsif ($arg eq 'issuer_name') {
            $encoded .= sprintf("Issuer Name: %s\n",
                                $self->issuer_name->$nameline());
        } elsif ($arg eq 'validity') {
            eval { use Time::localtime; };
            next if $@;

            my @expires = $self->expires;
            $encoded .= sprintf("Validity: Not Before - %s\n".
                                "          Not After  - %s\n",
                                ctime($expires[0]), ctime($expires[1]));
        } elsif ($arg eq 'fingerprint') {
            my $digest = $formats{'fingerprint'} || 'md5';
            my $fingerprint =
                join(':', $self->fingerprint(hex => $digest) =~ /..?/g);

            $encoded .= sprintf("%s-fingerprint: %s\n",
                                uc $digest, $fingerprint);
        } elsif ($arg eq 'bits') {
            $encoded .= sprintf("Public-key bits: %d\n", $self->bits);
        }
    }
    $encoded .= "\n" if $encoded;
    $encoded .= SiriSSL::PEM::_encode_cert($self);

    return $encoded;
}

sub _format {
    my $fmt = shift;
    if ($fmt =~ /^bin/i) {
        return sub { return shift @_ };
    } elsif ($fmt =~ /^hex/i) {
        return sub { return unpack("H*", shift @_) };
    } elsif ($fmt =~ /^b.*64$/) {
        return sub {
            eval { require MIME::Base64; };
            croak "MIME::Base64 module required to encode in base64" if $@;

            chomp(my $encoded = MIME::Base64::encode(shift @_));
            return $encoded;
        }
    }
    undef;
}

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    if(@_) {
        my $arg = shift;
        return $arg->dup() if(isa($arg, $class));
        return scalar(SiriSSL::PEM::decode_certs($arg));
    }

    return $class->_new();
}

sub cmp {
    my ($self, $other) = @_;
    $self->fingerprint() cmp $self->fingerprint;
}

sub fingerprint {
    my $self = shift;
    
    my %args;
    my @keys = ('digest', 'format');

    my $i = 0;
    while($i < @_) {
        my $arg = $_[$i];
        if(grep($arg eq $_, @keys)) {
            $args{$arg} = $_[$i+1];
            splice @_, $i, 2;
        } else {
            $i++;
        }
    }

    while (@_) {
        my $arg = shift;
        _format($arg) ?
            ($args{'format'} = $arg) :
                ($args{'digest'} = $arg);
    }
    
    $args{'format'} ||= 'hex';
    $args{'digest'} ||= 'sha';

    my $digest = $self->_fingerprint(lc $args{'digest'});
    my $formatter = _format($args{'format'});
    die "unknown format: $args{'format'}" unless $formatter;
    
    return $formatter->($digest);
}

sub check_time {
    my $self = shift;
    my $time = @_ ? shift : time;

    my ($notbefore, $notafter) = $self->expires;
    return ($time > $notbefore && $time < $notafter) ? 1 : 0;
}

package SiriSSL::X509_NAME;

use strict;
use overload
    '""' => \&shortline,
    'eq' => sub { return !($_[0]->oneline cmp $_[1]->oneline); },
    'cmp' => sub { return $_[0]->oneline cmp $_[1]->oneline; },
    'fallback' => 1;

use UNIVERSAL qw(isa);
use vars qw(%ALIASES $AUTOLOAD);

%ALIASES = (C => ['country', 'countryName'],
            CN => ['name', 'commonName'],
            ST => ['state', 'province', 'stateName', 'stateOrProvinceName'],
            L => ['loc', 'location', 'locality', 'localityName'],
            O => ['org', 'organization', 'organizatioName'],
            OU => ['unit', 'organizationUnit', 'organizationalUnit',
                   'organizationalUnitName'],
            Email => ['e-mail', 'e_mail'],

            # exotic, x509v3?
            T => ['title', 'titleName'],
            D => ['desc', 'description'],
            I => ['initials'],
            UID => ['uniqueIdentifier'],
            
            );

sub _lookup {
    my $attribute = shift;
    
    foreach (keys %ALIASES) {
        return $_ if (lc $attribute eq lc $_)
    }

    foreach (keys %ALIASES) {
        foreach my $alias (@{$ALIASES{$_}}) {
            return $_ if(lc $attribute eq lc $alias);
        }
    }
    die "unknown X509_NAME attribute or alias name ($attribute)";
}

# 'high-level' interface to the rest of the _parse_ lot.
sub _parse_line {
    my $line = shift;

    if($line =~ /^\/\w{1,5}=/) {
        return _parse_old_oneline($line);
    } elsif ($line =~ /^\w{1,5}=/) {
        return _parse_oneline($line);
    } else {
        return _parse_nameline($line);
    }
}

sub _parse_oneline {
    map(split(/=/), split(/, /, shift));
}

sub _parse_old_oneline {
    map(split(/=/), split(/\//, substr(shift, 1)));
}

sub _parse_nameline {
    my $name = shift;

    # this is pretty neat, it can match a name and or email out of
    # almost any format you might like to use.

    $name =~ s {
        \s*[:,-;]?\s*           # (optional) formatting border.
        [\<\[\(]?\s*            # open (optional) email delimiters.

        ([\w\d]+\@[\w\d]+[\w\d\.]*)     # email string ($1)

        \s*[\)\]\>]?\s*         # close (optional) email delimiters.
        } []x;

    my $email = $1;
    $name =~ s/^\s*(\S+.+\S+)\s*$/$1/;

    my %entries;
    $entries{CN} = $name if $name;
    $entries{Email} = $email if $email;

    return %entries;
}

sub match {
    my ($self, $match) = @_;

    foreach my $key (keys %{{$match->fetch}}) {
        my $self_value = $self->fetch($key);
        my $match_value = $match->fetch($key);

        if($match_value =~ s {^/(.*)/$} {$1}) {
            return 0 unless ($self_value =~ /$match_value/);
        } else {
            return 0 unless ($self_value eq $match_value);
        }
    }
    1;
}

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my $self = $class->_new();
    $self->store(@_);
    
    $self;
}

sub store {
    my $self = shift;
    if(@_ > 1) {
        my %entries = @_;
        while(my ($key, $value) = each %entries) {
            $key = _lookup($key);
            $self->_store($key => $value);
        }
    } elsif (@_) {
        my $arg = shift;

        return $self->store(%$arg)
            if ref($arg) eq 'HASH';

        return $self->store($arg->fetch())
            if isa($arg, ref $self);
        
        return $self->store(_parse_line($arg));
    }
    1;
}
        
sub fetch {
    my $self = shift;
    my @keys = @_;

    if(@keys) {
        my @values;
        foreach my $key (@keys) {
            $key = _lookup($key);
            push @values, $self->_fetch($key);
        }
        return wantarray ? @values : shift @values;
    } elsif(wantarray) {
        my %entries;
        foreach my $key ($self->_keys()) {
            $entries{$key} = $self->_fetch($key);
        }
        return %entries;
    } else {
        return $self->oneline;
    }
}

sub clear {
    my $self = shift;
    $self->_clear();
}

sub delete {
    my $self = shift;
    my @keys = @_;
    foreach my $key (@keys) {
        $key = _lookup($key);
        $self->_delete($key);
    }
    1;
}

sub shortline {
    my $self = shift;
    my ($name, $email) = $self->fetch("name", "email");

    if($name && $email) {
        return "$name <$email>";
    } elsif ($name) {
        return "$name";
    } elsif ($email) {
        return "$email";
    } else {
        return $self->oneline();
    }
}

*longline = \&oneline;
sub oneline {
    my $self = shift;
    my %entries = $self->fetch();
    return join(', ', map("$_=$entries{$_}", keys %entries));
}
*old_longline = \&old_oneline;

sub cmp {
    my ($self, $other) = @_;
    return $self->oneline cmp $other->oneline;
}

sub AUTOLOAD {
    my $method = $AUTOLOAD;

    $method =~ s/.*://;
    my $self = shift;
    
    $method = _lookup($method);
    $self->_store($method => shift(@_)) if @_;
    return $self->_fetch($method);
}

sub DESTROY {
}

package SiriSSL::CertStore;
use UNIVERSAL qw(isa);

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    
    my $self = bless {}, $class;

    $self->store(@_);
    $self;
}

sub dup {
    my $self = shift;
    my $clone = bless {}, ref($self);

    $clone->store($self->fetch);
    $clone;
}

sub parent {
    my $self = shift;
    if(@_) {
        my $parent = shift;
        $self->_parent_delete($self->fetch);
        $self->{'parent'} = $parent;
        $self->_parent_store($self->fetch);
    }
    return $self->{'parent'};
}

sub _parent_store {
    my $self = shift;
    my $parent = $self->parent;
    return $parent && $parent->_add_trusted_certs(@_);
}

sub _parent_delete {
    my $self = shift;
    my $parent = $self->parent;
    return $parent && $parent->_del_trusted_certs(@_);
}

sub exists {
    my $self = shift; my $other = shift;
    my $key = $other->fingerprint('binary', 'md5');
    return $self->{'certs'}->{$key};
}

sub fetch {
    my $self = shift;

    return values %{$self->{'certs'}} unless @_;
    
    my @results;

    my $arg = shift;
    if(isa($arg, 'SiriSSL::X509')) {
        @results = grep($self->exists($_), $arg, @_);
    } elsif (isa($arg, 'SiriSSL::X509_NAME')) {
        @results = grep($_->name->match($arg), values %{$self->{'certs'}});
    } else {
        @results = $self->fetch(SiriSSL::X509_NAME->new($arg, @_));
    }
    return wantarray ? @results : $results[0];
}

sub store {
    my $self = shift;
    my @certs = SiriSSL::PEM::decode_certs(@_);
    foreach my $cert (@certs) {
        my $key = $cert->fingerprint('binary', 'md5');
        $self->{'certs'}->{$key} = $cert;
    }
    $self->_parent_store(@certs);
    scalar(@certs);
}

sub delete {
    my $self = shift;
    my @args = @_;
    my @certs;
    foreach my $arg (@args) {
        $arg = SiriSSL::X509_NAME->new($arg)
            unless(isa($arg, 'SiriSSL::X509') ||
                   isa($arg, 'SiriSSL::X509_NAME'));
        push @certs, $self->fetch($arg);
    }
    foreach my $cert (@certs) {
        my $key = $cert->fingerprint('binary', 'md5');
        delete $self->{'certs'}->{$key};
    }
    $self->_parent_delete(@certs);
    scalar(@certs);
}

sub clear {
    my $self = shift;
    $self->_parent_delete($self->fetch);
    $self->{'certs'} = {};
}

sub DESTROY {
    my $self = shift;
    print "Destroying $self!\n";
}

1;
__END__
