#include "SiriSSL.h"
#include <ctype.h>

#define X509_name               X509_get_subject_name
#define X509_issuer_name        X509_get_issuer_name

#define X509_bits               X509__bits
#define X509_check_signature    X509__check_signature

int
X509_V_ERR_CONSTANT(name)
        char *name;
{
        errno = 0;
        switch(*name) {

	case 'A':
                if(strEQ(name, "APPLICATION_VERIFICATION"))
                        return X509_V_ERR_APPLICATION_VERIFICATION;
		break;

	case 'C':
                if(strEQ(name, "CERT_CHAIN_TOO_LONG"))
                        return X509_V_ERR_CERT_CHAIN_TOO_LONG;
                if(strEQ(name, "CERT_HAS_EXPIRED"))
                        return X509_V_ERR_CERT_HAS_EXPIRED;
                if(strEQ(name, "CERT_NOT_YET_VALID"))
                        return X509_V_ERR_CERT_NOT_YET_VALID;

                if(strEQ(name, "CERT_REVOKED"))
                        return X509_V_ERR_CERT_REVOKED;
                if(strEQ(name, "CERT_SIGNATURE_FAILURE"))
                        return X509_V_ERR_CERT_SIGNATURE_FAILURE;
                if(strEQ(name, "CRL_HAS_EXPIRED"))
                        return X509_V_ERR_CRL_HAS_EXPIRED;
                if(strEQ(name, "CRL_NOT_YET_VALID"))
                        return X509_V_ERR_CRL_NOT_YET_VALID;
                if(strEQ(name, "CRL_SIGNATURE_FAILURE"))
                        return X509_V_ERR_CRL_SIGNATURE_FAILURE;
		break;

	case 'D':
                if(strEQ(name, "DEPTH_ZERO_SELF_SIGNED_CERT"))
                        return X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT;
		break;

	case 'E':
                if(strEQ(name, "ERROR_IN_CERT_NOT_AFTER_FIELD"))
                        return X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD;
                if(strEQ(name, "ERROR_IN_CERT_NOT_BEFORE_FIELD"))
                        return X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD;
                if(strEQ(name, "ERROR_IN_CRL_LAST_UPDATE_FIELD"))
                        return X509_V_ERR_ERROR_IN_CRL_LAST_UPDATE_FIELD;
                if(strEQ(name, "ERROR_IN_CRL_NEXT_UPDATE_FIELD"))
                        return X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD;
		break;

	case 'O':
                if(strEQ(name, "OUT_OF_MEM"))
                        return X509_V_ERR_OUT_OF_MEM;
		break;

	case 'S':
                if(strEQ(name, "SELF_SIGNED_CERT_IN_CHAIN"))
                        return X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN;
		break;

	case 'U':
                if(strEQ(name, "UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY"))
                        return X509_V_ERR_UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY;
                if(strEQ(name, "UNABLE_TO_DECRYPT_CERT_SIGNATURE"))
                        return X509_V_ERR_UNABLE_TO_DECRYPT_CERT_SIGNATURE;
                if(strEQ(name, "UNABLE_TO_DECRYPT_CRL_SIGNATURE"))
                        return X509_V_ERR_UNABLE_TO_DECRYPT_CRL_SIGNATURE;
                if(strEQ(name, "UNABLE_TO_GET_CRL"))
                        return X509_V_ERR_UNABLE_TO_GET_CRL;
                if(strEQ(name, "UNABLE_TO_GET_ISSUER_CERT"))
                        return X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT;
                if(strEQ(name, "UNABLE_TO_GET_ISSUER_CERT_LOCALLY"))
                        return X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY;
                if(strEQ(name, "UNABLE_TO_VERIFY_LEAF_SIGNATURE"))
                        return X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE;
		break;
        }
        errno = EINVAL;
        return 0;
}


int
X509_V_CONSTANT(name)
        char *name;
{
        errno = 0;

        if(strEQ(name, "OK"))
                return X509_V_OK;

        if(strnEQ(name, "ERR_", 4)) {
                int value;
                value = X509_V_ERR_CONSTANT(name + 4);
                if(!errno) return value;
        }

        errno = EINVAL;
        return 0;
}

EVP_MD *__get_digest(SV *digest_name)
{
        EVP_MD *digest;

        if(SvPOK(digest_name)) {
                if(digest = EVP_get_digestbyname(SvPVX(digest_name)))
                        return digest;

                croak("Unknown digest algorithm: %s", SvPVX(digest_name));
        }

        croak("digest algorithm name is not a string!");
}

time_t __time_asn2unix(ASN1_TIME *asn1_time)
{
        char *  cp;
        int     len;

        int             seconds, minutes, hours, month_day, month, year;
        struct tm       parsed_time;
                
        cp = asn1_time->data;
        len = asn1_time->length;

        if(asn1_time->type == V_ASN1_UTCTIME) {
                if(len < 10)
                        return (time_t)-1;

                sscanf(cp, "%02d%02d%02d%02d%02d", &year, &month, &month_day,
                                                   &hours, &minutes);
                if(year < 50) year += 100;

                seconds = 0;
                if(isdigit(cp[10]) && isdigit(cp[11]))
                        sscanf(&cp[10], "%02d", &seconds);
        } else if (asn1_time->type == V_ASN1_GENERALIZEDTIME) {
                if(len < 12)
                        return (time_t)-1;

                sscanf(cp, "%04d%02d%02d%02d%02d", &year, &month, &month_day,
                                                   &hours, &minutes);
                year -= 1900;
                seconds = 0;
                if(isdigit(cp[12]) && isdigit(cp[13]))
                        sscanf(&cp[12], "%02d", &seconds);
        }

        parsed_time.tm_sec = seconds;
        parsed_time.tm_min = minutes;
        parsed_time.tm_mon = month - 1;
        parsed_time.tm_hour = hours;
        parsed_time.tm_mday = month_day;
        parsed_time.tm_year = year;

        return mktime(&parsed_time);
}

static int
X509__bits(X509 *x509)
{
        int bits;
        EVP_PKEY *pkey = X509_get_pubkey(x509);

        if(!pkey) return 0;
        bits = EVP_PKEY_bits(pkey);
        EVP_PKEY_free(pkey);

        return bits;
}

static int
X509__check_signature(subject, issuer)
        X509 *subject, *issuer;
{
        EVP_PKEY *pkey = X509_get_pubkey(issuer);
        return (pkey && X509_verify(subject, pkey) > 0) ? 1 : 0;
}

static SV *
X509__fingerprint(cert, md)
        X509 *          cert;
        EVP_MD *        md;
{
        int     digest_len;
        SV *    digest;

        SV_ALLOC(digest, DIGEST_MAX_LEN);        

        X509_digest(cert, md, SvPVX(digest), &digest_len);
        SvCUR_set(digest, digest_len);

        return digest;
}

MODULE = SiriSSL::X509  PACKAGE = SiriSSL::X509         PREFIX = X509_

X509 *
X509__new(class)
                SV *class
        CODE:
                RETVAL = X509_new();
        OUTPUT:
                RETVAL

X509 *
X509_dup(cert)
        X509 *cert

void
X509_expires(cert)
                X509 *cert
        PPCODE:
                if(GIMME_V == G_ARRAY) {
                        XPUSH_NV(__time_asn2unix(X509_get_notBefore(cert)));
                        XPUSH_NV(__time_asn2unix(X509_get_notAfter(cert)));
                } else {
                        XPUSH_NV(__time_asn2unix(X509_get_notAfter(cert)));
                }


SV *
X509__fingerprint(cert, ...)
                X509 *cert
        PREINIT:
                EVP_MD  *digest;
        CODE:
                if(items > 1)
                        digest = __get_digest(ST(1));
                else
                        digest = X509_DEFAULT_DIGEST;

                RETVAL = X509__fingerprint(cert, digest);
        OUTPUT:
                RETVAL


int
X509_bits(x509)
        X509 *x509


X509_NAME *
X509_name(x509)
        X509 *x509


X509_NAME *
X509_issuer_name(x509)
        X509 *x509

int
X509_check_private_key(x509, pkey)
        X509 *          x509
        EVP_PKEY *      pkey


char *
X509_cert_error_as_string(class, verify_result)
        SV *class
        int verify_result
        CODE:
                RETVAL = X509_verify_cert_error_string(verify_result);
        OUTPUT:
                RETVAL

void
X509_DESTROY(...)
        PREINIT:
                X509 *x509;
        CODE:
                CLASS_IN(ST(0), x509, X509 *, SiriSSL::X509);
                printd("Destroying x509 perl object (0x%x): %d references\n",
                      (unsigned long)x509, x509->references);
                X509_free(x509);                

EVP_PKEY *
X509_get_pubkey(x509)
        X509 *x509

int
X509_check_signature(subject, issuer)
        X509 *subject
        X509 *issuer
