#include "SiriSSL.h"

static int
__txt2nid(char *txt)
{
        int nid = OBJ_txt2nid(txt);
        if(nid == NID_undef)
                croak("Invalid X509_NAME attribute name (%s)", txt);
        return nid;                
}

static char *
X509_NAME_ENTRY_key(X509_NAME_ENTRY *entry)
{
        return OBJ_nid2sn(OBJ_obj2nid(X509_NAME_ENTRY_get_object(entry)));
}

static SV *
X509_NAME_ENTRY_value(entry)
        X509_NAME_ENTRY *entry;
{
        ASN1_STRING *   data = X509_NAME_ENTRY_get_data(entry);

        /* TODO: check out all those conversions done in X509_NAME_oneline,
           on the entry value, maybe we need to do some of these too. */

        return newSVpvn(ASN1_STRING_data(data), ASN1_STRING_length(data));
}

static int
X509_NAME_ENTRY_value_type(nid, cp, len)
        int     nid;
        char *  cp;
        int     len;
{
        int type;

        type = ASN1_PRINTABLE_type(cp, len);

	if (nid == NID_pkcs9_emailAddress)
		type = V_ASN1_IA5STRING;

	if ((nid == NID_commonName) && (type == V_ASN1_IA5STRING))
		type = V_ASN1_T61STRING;

	if ((nid == NID_pkcs9_challengePassword) && (type == V_ASN1_IA5STRING))
		type = V_ASN1_T61STRING;

	if (nid == NID_pkcs9_unstructuredName)
		type = V_ASN1_IA5STRING;

	return type;
}


static SV *
X509_NAME__fetch(name, key)
        X509_NAME *     name;
        char *          key;
{
        int                     index;

        X509_NAME_ENTRY *       entry;
        char *                  entry_nid;

        int                     key_nid = __txt2nid(key);

        if((index = X509_NAME_get_index_by_NID(name, key_nid, -1)) < 0)
                return &sv_undef;

        entry = X509_NAME_get_entry(name, index);
        return X509_NAME_ENTRY_value(entry);
}

static void
X509_NAME__store(name, key, value)
        X509_NAME *     name;
        char *          key;
        SV *            value;
{
        X509_NAME_ENTRY *       entry;

        int index, type, nid;

        char *  cp;
        int     len;

        cp = SvPOK(value) ?
                SvPV(value, len) :
                SvPV_force(sv_2mortal(newSVsv(value)), len);
        
        nid = __txt2nid(key);
        type = X509_NAME_ENTRY_value_type(nid, cp, len);

        entry = X509_NAME_ENTRY_create_by_NID(NULL, nid, type, cp, len);

        index = X509_NAME_get_index_by_NID(name, nid, -1);
        if(index < 0)
                sk_push(name->entries, (char *)entry);
        else {
                X509_NAME_ENTRY *       old_entry;
                STACK *                 entries;

                entries = name->entries;

                old_entry = (X509_NAME_ENTRY *)sk_delete(entries, index);
                sk_insert(entries, (char *)entry, index);

                X509_NAME_ENTRY_free(old_entry);
        }
        name->modified = 1;
}
        
static int
X509_NAME__delete(name, key)
        X509_NAME *     name;
        char *          key;
{
        int index;
        int key_nid = __txt2nid(key);

        if((index = X509_NAME_get_index_by_NID(name, key_nid, -1)) < 0)
                return 0;
        return X509_NAME_delete_entry(name, index) ? 1 : 0;
}

static void
X509_NAME__clear(name)
        X509_NAME *     name;
{
        int count;
        while(count = X509_NAME_entry_count(name))
                X509_NAME_delete_entry(name, count - 1);
}


#define XPUSH_X509_NAME_KEYS(name) \
do { \
        int i; \
        char *key; \
        \
        X509_NAME_ENTRY *entry; \
        \
        for(i = 0; i < X509_NAME_entry_count(name); i++) { \
                entry = X509_NAME_get_entry(name, i); \
                key = X509_NAME_ENTRY_key(entry); \
                XPUSH_PV(key, 0); \
        } \
} while(0)

MODULE = SiriSSL::X509_NAME  PACKAGE = SiriSSL::X509_NAME   PREFIX = X509_NAME_


void
X509_NAME__keys(name)
                X509_NAME *     name
        PPCODE:
                XPUSH_X509_NAME_KEYS(name);


void
X509_NAME__clear(name)
                X509_NAME *     name

int
X509_NAME__delete(name, key)
                X509_NAME *     name
                char *          key

void
X509_NAME__store(name, key, value)
                X509_NAME *     name
                char *          key
                SV *            value

SV *
X509_NAME__fetch(name, key)
                X509_NAME *     name
                char *          key


X509_NAME *
X509_NAME__new(class)
                SV *    class
        CODE:
                RETVAL = X509_NAME_new();
        OUTPUT:
                RETVAL

long
X509_NAME_hash(name)
        X509_NAME *name

SV *
X509_NAME_old_oneline(name)
        X509_NAME *name
        PREINIT:
                char *oneline;
        CODE:
                oneline = X509_NAME_oneline(name, NULL, 0);
                RETVAL = newSVpv(oneline, 0);
                free(oneline);
        OUTPUT:
                RETVAL

X509_NAME *
X509_NAME_dup(name)
        X509_NAME *name

