#include "SiriSSL.h"

#define PERL_VERIFY_CB  10
#define PERL_TMPRSA_CB  11
#define PERL_TMPDH_CB   12
#define PERL_INFO_CB    13

#define SSL_EXT_IO_RETVAL       14
#define PERL_CERT_STORE         15

/* superglue the SiriSSL::CertStore pure perl object to our interface. */

SV *
CertStore_new(parent)
        SSL_CTX *       parent;
{
        SV *cert_store = perl_eval_pv("SiriSSL::CertStore->new()" , 1);
        CertStore_parent(cert_store, parent);

        return cert_store;
}

void
CertStore_parent(cert_store, parent)
        SV *            cert_store;
        SSL_CTX *       parent;
{
        dSP;

        REF_INC(parent, SSL_CTX);
        
        PUSHMARK(sp);
        XPUSHs(cert_store);
        XPUSH_OBJ(parent, SiriSSL::SSL_CTX);
        PUTBACK;

        perl_call_method("parent", G_DISCARD);
}

void *
sv_unref_ptr(SV *sv_ptr, char *class)
{
        void *obj;
        
        if(sv_derived_from(sv_ptr, class))
                obj = (void *)SvIV((SV *)SvRV(sv_ptr));
        else
                croak("expecting a class %s object", class);

        return obj;
}

void
ex_store(CRYPTO_EX_DATA *ex_data, char *extension, int idx, void *val)
{
        int ok = CRYPTO_set_ex_data(ex_data, idx, val);
        if(!ok)
                croak("can't store extension (%s): %s", extension,
                      SSL_strerror(SSL_errno()));
}

void *
ex_fetch(CRYPTO_EX_DATA *ex_data, char *extension, int idx, int demand)
{
        void *val = CRYPTO_get_ex_data(ex_data, idx);
        if(demand && !val)
                croak("can't fetch extension (%s)", extension);
        return val;
}

/**************************
 *** SSL API extensions ***
 **************************/

/* TODO:
   Resolve memory leak dereferencing issue:

   When OpenSSL objects are destroyed they do not dereference their
   by-extension stored perl objects.
*/

void
X509_STORE_set_perl_verify_callback(X509_STORE *cert_store, SV *callback)
{
        SV *sv = (SV *) EXTENSION_FETCH(cert_store, PERL_VERIFY_CB);

        SV_SETSV(sv, callback);
        EXTENSION_STORE(cert_store, PERL_VERIFY_CB, sv);
}

SV *
X509_STORE_get_perl_verify_callback(X509_STORE *cert_store)
{
        return (SV *)  EXTENSION_DEMAND(cert_store, PERL_VERIFY_CB);
}

void
SSL_set_io_retval(SSL *ssl, int io_retval)
{
        SV *sv = (SV *) EXTENSION_FETCH(ssl, SSL_EXT_IO_RETVAL);

        SV_SETIV(sv, io_retval);
        EXTENSION_STORE(ssl, SSL_EXT_IO_RETVAL, sv);
}

int
SSL_get_io_retval(SSL *ssl)
{
        SV *sv = (SV *) EXTENSION_FETCH(ssl, SSL_EXT_IO_RETVAL);

        if(sv == (SV *)NULL)
        /* positive default to satisfy the error routines */
                return 1;
        else
                return SvIV(sv);
}

/* memory leak, noone will dereference this hybrid ssl-perl object */
void
SSL_CTX_set_perl_cert_store(SSL_CTX *ssl_ctx, SV *cert_store)
{
        SV *tmp = (SV *)EXTENSION_FETCH(ssl_ctx, PERL_CERT_STORE);
        SV_SETSV(tmp, cert_store);

        EXTENSION_STORE(ssl_ctx, PERL_CERT_STORE, tmp);
}

SV *
SSL_CTX_get_perl_cert_store(SSL_CTX *ssl_ctx)
{
        return (SV *)EXTENSION_FETCH(ssl_ctx, PERL_CERT_STORE);
}

void
SSL_CTX_set_perl_tmp_rsa_callback(SSL_CTX *ssl_ctx, SV *cb)
{
        SV *callback = EXTENSION_FETCH(ssl_ctx, PERL_TMPRSA_CB);

        SV_SETSV(callback, cb);
        EXTENSION_STORE(ssl_ctx, PERL_TMPRSA_CB, callback);
}

SV *
SSL_CTX_get_perl_tmp_rsa_callback(SSL_CTX *ssl_ctx)
{
        return (SV *)EXTENSION_DEMAND(ssl_ctx, PERL_TMPRSA_CB);
}

void
SSL_CTX_set_perl_tmp_dh_callback(SSL_CTX *ssl_ctx, SV *cb)
{
        SV *callback = EXTENSION_FETCH(ssl_ctx, PERL_TMPDH_CB);

        SV_SETSV(callback, cb);
        EXTENSION_STORE(ssl_ctx, PERL_TMPDH_CB, callback);
}

SV *
SSL_CTX_get_perl_tmp_dh_callback(SSL_CTX *ssl_ctx)
{
        return (SV *)EXTENSION_DEMAND(ssl_ctx, PERL_TMPDH_CB);
}

void
SSL_CTX_set_perl_info_callback(SSL_CTX *ssl_ctx, SV *cb)
{
        SV *callback = EXTENSION_FETCH(ssl_ctx, PERL_INFO_CB);

        SV_SETSV(callback, cb);
        EXTENSION_STORE(ssl_ctx, PERL_INFO_CB, callback);
}

SV *
SSL_CTX_get_perl_info_callback(SSL_CTX *ssl_ctx)
{
        return (SV *)EXTENSION_DEMAND(ssl_ctx, PERL_INFO_CB);
}

/* I don't like to doing it this way either */

static SV *GLOBAL_passwd_callback = NULL;

void
GLOBAL_set_perl_passwd_callback(SV *cb)
{
        if(cb == &sv_undef)
                GLOBAL_passwd_callback = NULL;
        else
                SV_SETSV(GLOBAL_passwd_callback, cb);
}

SV *
GLOBAL_get_perl_passwd_callback(int demand)
{
        if(demand && !GLOBAL_passwd_callback)
                croak("perl_passwd_callback hasn't been defined!");
        return GLOBAL_passwd_callback;
}

/*************************
 *** Callback wrappers ***
 *************************/

int     perl_verify_callback(int ok, X509_STORE_CTX *store_ctx)
{
        dSP;
        int count;
        SV *callback = X509_STORE_get_perl_verify_callback(store_ctx->ctx);

        ENTER;
        SAVETMPS;

        PUSHMARK(sp);
        XPUSH_IV(ok);
        XPUSH_OBJ(store_ctx, SiriSSL::X509_STORE_CTX);
        PUTBACK;

        count = perl_call_sv(callback, G_SCALAR);

        SPAGAIN;
        if(count != 1)
                croak("perl_verify_callback: "
                      "callback returned a list instead of a scalar!");
        ok = POPi;

        PUTBACK;
        FREETMPS;
        LEAVE;

        return ok;
}

void    perl_info_callback(SSL *ssl, int where, int ret)
{
        dSP;
        SV *callback = SSL_CTX_get_perl_info_callback(SSL_get_SSL_CTX(ssl));

        ENTER;
        SAVETMPS;

        PUSHMARK(sp);
        XPUSH_OBJ(ssl, SiriSSL::SSL);
        XPUSH_IV(where);
        XPUSH_IV(ret);
        PUTBACK;
        
        perl_call_sv(callback, G_DISCARD);
        
        FREETMPS;
        LEAVE;
}

RSA *   perl_tmp_rsa_callback(SSL *ssl, int export, int keylen)
{
        dSP;
        int count;
        RSA *rsa;
        SV *callback = SSL_CTX_get_perl_tmp_rsa_callback(SSL_get_SSL_CTX(ssl));

        ENTER;
        SAVETMPS;

        PUSHMARK(sp);
        XPUSH_OBJ(ssl, SiriSSL::SSL);
        XPUSH_IV(export);
        XPUSH_IV(keylen);
        PUTBACK;
        
        count = perl_call_sv(callback, G_SCALAR);

        SPAGAIN;

        if(count != 1) 
                croak("perl_tmp_rsa_callback: "
                      "callback returned a list instead of a scalar!");

        rsa = (RSA *)POP_OBJ(SiriSSL::RSA);
        
        PUTBACK;
        FREETMPS;
        LEAVE;

        return rsa;
}


DH *   perl_tmp_dh_callback(SSL *ssl, int export, int keylen)
{
        dSP;
        int count;
        DH *dh;
        SV *callback = SSL_CTX_get_perl_tmp_dh_callback(SSL_get_SSL_CTX(ssl));

        ENTER;
        SAVETMPS;

        PUSHMARK(sp);
        XPUSH_OBJ(ssl, SiriSSL::SSL);
        XPUSH_IV(export);
        XPUSH_IV(keylen);
        PUTBACK;
        
        count = perl_call_sv(callback, G_SCALAR);

        SPAGAIN;

        if(count != 1) 
                croak("perl_tmp_dh_callback: "
                      "callback returned a list instead of a scalar!");

        dh = (DH *)POP_OBJ(SiriSSL::DH);
        
        PUTBACK;
        FREETMPS;
        LEAVE;

        return dh;
}

int     perl_passwd_callback(char *buf, int buf_len, int verify)
{
        dSP;
        int count, len;
        char *passwd;
        SV *callback = GLOBAL_get_perl_passwd_callback(1);

        ENTER;
        SAVETMPS;

        PUSHMARK(sp);
        XPUSH_IV(verify);
        PUTBACK;
        
        count = perl_call_sv(callback, G_SCALAR);

        SPAGAIN;

        if(count != 1)
                croak("perl_passwd_callback: "
                      "callback returned a list instead of a scalar");

        passwd = POPp;
        if(strlen(passwd) > buf_len)
                len = buf_len;
        else
                len = strlen(passwd);
        snprintf(buf, buf_len, "%s", passwd);
        buf[buf_len - 1] = '\0';
        
        PUTBACK;
        FREETMPS;
        LEAVE;

        return len;
}
